@@ -1,103 +1,10 @@
 /* es.h -- definitions for higher order shell */
 
-#include <stddef.h>
-#include <stdarg.h>
-#include <string.h>
-#include <errno.h>
-#include <setjmp.h>
-#include <signal.h>
-
 #include "config.h"
+#include "stdenv.h"
 
 /*
- * augmentations to standard c environment
- */
-
-#undef FALSE
-#undef TRUE
-typedef enum { FALSE, TRUE } Boolean;
-
-#ifndef	offsetof
-#define	offsetof(t, m)	((size_t) (((char *) &((t *) 0)->m) - (char *)0))
-#endif
-
-#define	streq(s, t)		(strcmp(s, t) == 0)
-#define	strneq(s, t, n)		(strncmp(s, t, n) == 0)
-#define arraysize(a)		((int)(sizeof(a)/sizeof(*a)))
-#define	memzero(dest, count)	memset(dest, 0, count)
-
-#if __GNUC__
-typedef volatile void noreturn;
-#else
-typedef void noreturn;
-#endif
-
-#if GCDEBUG && NeXT
-#define	ABORT()	asm("trap #15")	/* (*(char *) 0 = 'c'); */
-#else
-#define	ABORT()	abort()
-#endif
-
-#if ASSERTIONS
-#define assert(expr) \
-	do { \
-		if (!(expr)) { \
-			eprint("%s:%d: assertion failed (%s)\n", \
-					__FILE__, __LINE__, #expr); \
-			ABORT(); \
-		} \
-	} while (0)
-#else
-#define assert(ignore) do ; while (0)
-#endif
-
-#ifndef	EOF
-#define	EOF	(-1)
-#endif
-
-#ifndef	NULL
-#define	NULL	((void *) 0)
-#endif
-
-extern noreturn abort(void);
-
-
-/*
- * system protos used; do something about this
- */
-
-/* stdlib */
-extern noreturn exit(int);
-extern long strtol(const char *num, char **end, int base);
-#define	atoi(s)	strtol(s, NULL, 0)
-
-/* string */
-extern void *memcpy(void *dst, const void *src, size_t n);
-extern void *memset(void *dst, int c, size_t n);
-
-/* syscalls */
-extern int read(int fd, void *buf, size_t n);
-extern int write(int fd, const void *buf, size_t n);
-extern int dup(int fd);
-extern int dup2(int srcfd, int dstfd);
-extern void close(int fd);
-extern int getpagesize(void);
-extern int pipe(int[2]);
-extern int getpid(void);
-
-
-/*
- * redirection types:  these would be in token.h, but they are also used for
- * eopen() from open.c
- */
-
-typedef enum {
-	rOpen, rCreate, rAppend, rDup, rClose, rPipe, rHeredoc, rHerestring
-} RedirKind;
-
-
-/*
- * the fundamental es data structures.  
+ * the fundamental es data structures.
  */
 
 typedef struct Tree Tree;
@@ -125,6 +32,7 @@ struct Binding {
 struct Closure {
 	Binding	*binding;
 	Tree *tree;
+	int mark;
 };
 
 
@@ -133,8 +41,9 @@ struct Closure {
  */
 
 typedef enum {
-	nAssign, nCall, nConcat, nFor, nLambda, nLet, nList, nLocal,
-	nMatch, nPrim, nQword, nRec, nThunk, nVar, nVarsub, nWord
+	nAssign, nCall, nClosure, nConcat, nFor, nLambda, nLet, nList, nLocal,
+	nMatch, nPrim, nQword, nThunk, nVar, nVarsub, nWord,
+	nRedir, nPipe		/* only appear during construction */
 } NodeKind;
 
 struct Tree {
@@ -158,7 +67,7 @@ struct StrList {
 };
 
 typedef struct {
-	int len;
+	int alloclen, count;
 	char *vector[1];
 } Vector;			/* environment or arguments */
 
@@ -167,16 +76,36 @@ typedef struct {
  * our programming environment
  */
 
+
 /* main.c */
 
-extern Boolean interactive;		/* -i or isatty(input) */
 extern Boolean loginshell;		/* -l or $0[0] == '-' */
 extern Boolean noexecute;		/* -n */
-extern Boolean verbose;			/* -v */
-extern Boolean printcmds;		/* -x */
+extern Boolean exitonfalse;		/* -e */
 
-extern Boolean gcdebug;			/* -G */
-extern Boolean bugme;			/* -B */
+#if GCVERBOSE
+extern Boolean gcverbose;		/* -G */
+#endif
+#if GCINFO
+extern Boolean gcinfo;			/* -I */
+#endif
+
+
+/* fd.c */
+
+extern void mvfd(int old, int new);
+extern int newfd(void);
+
+#define	UNREGISTERED	(-999)
+extern void registerfd(int *fdp, Boolean closeonfork);
+extern void unregisterfd(int *fdp);
+extern void releasefd(int fd);
+extern void closefds(void);
+
+extern int fdmap(int fd);
+extern int defer_mvfd(Boolean parent, int old, int new);
+extern int defer_close(Boolean parent, int fd);
+extern void undefer(int ticket);
 
 
 /* term.c */
@@ -196,26 +125,26 @@ extern List *listcopy(List *list);
 extern int length(List *list);
 extern List *listify(int argc, char **argv);
 extern Term *nth(List *list, int n);
+extern List *sortlist(List *list);
 
 
 /* tree.c */
 
-extern Tree *mk(NodeKind, ...);
-extern Tree *revtree(Tree *tree);
+extern Tree *mk(NodeKind VARARGS);
 
 
 /* closure.c */
 
 extern Closure *mkclosure(Tree *tree, Binding *binding);
 extern Closure *extractbindings(Tree *tree);
-extern Binding *bind(char *name, List *defn, Binding *next);
+extern Binding *mkbinding(char *name, List *defn, Binding *next);
 
 
 /* eval.c */
 
-extern List *walk(Tree *tree, Binding *binding, Boolean parent);
-extern List *eval(List *list, Binding *binding, Boolean parent);
-extern List *eval1(Term *term, Boolean parent);
+extern List *walk(Tree *tree, Binding *binding, Boolean parent, Boolean exitonfalse);
+extern List *eval(List *list, Binding *binding, Boolean parent, Boolean exitonfalse);
+extern List *eval1(Term *term, Boolean parent, Boolean exitonfalse);
 
 
 /* glom.c */
@@ -237,15 +166,16 @@ extern Boolean listmatch(List *subject, List *pattern, StrList *quote);
 
 /* var.c */
 
-extern void initvars(char **envp, const char *initial, Boolean protected);
+extern void initvars(char **envp, Boolean protected);
 extern char *varname(List *);
-extern List *varlookup(char *, Binding *);
+extern List *varlookup(const char *, Binding *);
 extern List *varlookup2(char *name1, char *name2);
 extern void vardef(char *, Binding *, List *);
 extern void varpush(char *, List *);
 extern void varpop(char *);
 extern Vector *mkenv(void);
 extern void noexport(char *name);
+extern Boolean isnoexport(const char *name);
 
 
 /* status.c */
@@ -260,13 +190,13 @@ extern void printstatus(int pid, int status);
 /* proc.c */
 
 extern Boolean hasforked;
-extern int efork(Boolean parent, Boolean continuing, Boolean background);
-extern int ewaitfor(int pid);
-
+extern int efork(Boolean parent, Boolean background);
+extern int ewaitfor2(int pid, void *rusage);
+#define	ewaitfor(pid)	ewaitfor2(pid, NULL)
 
 /* which.c */
 
-extern char *which(char *name, Boolean verbose);
+extern char *which(char *name);
 
 
 /* dict.c */
@@ -282,22 +212,20 @@ extern void *dictget2(Dict *dict, const char *name1, const char *name2);
 /* conv.c */
 
 extern void initconv(void);
-extern int defaultfd(RedirKind op);
 
 
 /* print.c -- see print.h for more */
 
-extern int print(const char *fmt,...);
-extern int eprint(const char *fmt,...);
-extern int fprint(int fd, const char *fmt,...);
-extern noreturn panic(const char *fmt, ...);
-extern void debug(const char *fmt, ...);
+extern int print(const char *fmt VARARGS);
+extern int eprint(const char *fmt VARARGS);
+extern int fprint(int fd, const char *fmt VARARGS);
+extern noreturn panic(const char *fmt VARARGS);
 
 
 /* str.c */
 
-extern char *str(const char *fmt, ...);		/* create a gc space string by printing */
-extern char *mprint(const char *fmt, ...);	/* create an ealloc space string by printing */
+extern char *str(const char *fmt VARARGS);	/* create a gc space string by printing */
+extern char *mprint(const char *fmt VARARGS);	/* create an ealloc space string by printing */
 extern StrList *mkstrlist(char *, StrList *);
 
 
@@ -305,6 +233,7 @@ extern StrList *mkstrlist(char *, StrList *);
 
 extern Vector *mkvector(int n);
 extern Vector *vectorize(List *list);
+extern void sortvector(Vector *v);
 
 
 /* util.c */
@@ -314,31 +243,33 @@ extern void uerror(char *msg);
 extern void *ealloc(size_t n);
 extern void *erealloc(void *p, size_t n);
 extern void efree(void *p);
-extern void ewrite(int fd, char *s, size_t n);
+extern char *strdup(const char *s);
+extern void ewrite(int fd, const char *s, size_t n);
 extern long eread(int fd, char *buf, size_t n);
-extern int mvfd(int old, int new);
-extern int cpfd(int old, int new);
 extern Boolean isabsolute(char *path);
 
 
 /* input.c */
 
 extern char *prompt, *prompt2;
-extern List *runstring(const char *str);
+extern Tree *parse(char *esprompt1, char *esprompt2);
 extern Tree *parsestring(const char *str);
-extern List *runfd(int fd);
 extern void sethistory(char *file);
-
-extern int gchar(void);
-extern void ugchar(int c);
-extern void flushu(void);
-extern void closefds(void);
+extern Boolean isinteractive(void);
 extern void initinput(void);
 
+extern List *runfd(int fd, const char *name, int flags);
+extern List *runstring(const char *str, const char *name, int flags);
+
+#define	run_interactive		 1	/* -i or $0[0] = '-' */
+#define	run_noexec		 2	/* -n */
+#define	run_echoinput		 4	/* -v */
+#define	run_printcmds		 8	/* -x */
+#define	run_lisptrees		16	/* -L and defined(LISPTREES) */
 
 /* prim.c */
 
-extern List *prim(char *s, List *list, Boolean parent);
+extern List *prim(char *s, List *list, Boolean parent, Boolean exitonfalse);
 extern void initprims(void);
 
 
@@ -347,31 +278,28 @@ extern void initprims(void);
 extern void startsplit(const char *sep, Boolean coalesce);
 extern void splitstring(char *in, size_t len, Boolean endword);
 extern List *endsplit(void);
+extern List *fsplit(const char *sep, List *list);
 
 
 /* signal.c */
 
-#if USE_SIG_ATOMIC_T
-#define SIG_ATOMIC_T sig_atomic_t
-#else
-#define SIG_ATOMIC_T int
-#endif
-
-#define SIGCHK() sigchk()
-extern volatile SIG_ATOMIC_T slow, interrupted;
+#define	SIGCHK() sigchk()
+extern Atomic slow, interrupted;
 extern jmp_buf slowlabel;
 extern Boolean sigint_newline;
 extern void catcher(int sig);
 extern void sigchk(void);
-extern void (*esignal(int sig, void (*h)(int)))(int);
+extern sigresult (*esignal(int sig, sigresult (*h)(int)))(int);
 extern void initsignals(Boolean allowdumps);
 extern Boolean issilentsignal(List *e);
 extern void setsigdefaults(Boolean background);
+extern void trapsignals(Boolean sigs[]);
 
 
 /* open.c */
 
-extern int eopen(char *name, RedirKind k);
+typedef enum { oOpen, oCreate, oAppend } OpenKind;
+extern int eopen(char *name, OpenKind k);
 
 
 /* version.c */
@@ -382,7 +310,7 @@ extern const char * const version;
 /* gc.c -- see gc.h for more */
 
 typedef struct Tag Tag;
-#define	gcnew(type)	((type *) gcalloc(sizeof (type), &type ## Tag))
+#define	gcnew(type)	((type *) gcalloc(sizeof (type), &(CONCAT(type,Tag))))
 
 extern void *gcalloc(size_t n, Tag *t);		/* allocate n with collection tag t */
 extern char *gcdup(const char *s);		/* copy a 0-terminated string into gc space */
@@ -392,10 +320,7 @@ extern void initgc(void);			/* must be called at the dawn of time */
 extern void gc(void);				/* provoke a collection, if enabled */
 extern void gcenable(void);			/* enable collections */
 extern void gcdisable(size_t);			/* disable collections, collect first if space needed */
-
-#ifndef GARBAGE_COLLECTOR
-extern int gcblocked;
-#endif
+extern Boolean gcisblocked();			/* is collection disabled? */
 
 
 /*
@@ -411,27 +336,34 @@ struct Root {
 extern Root *rootlist;
 
 #define	Ref(t, v, init) \
-	do { \
+	if (0) ; else { \
 		t v = init; \
-		Root (__root__ ## v); \
-		(__root__ ## v).p = (void **) &v; \
-		(__root__ ## v).next = rootlist; \
-		rootlist = &(__root__ ## v)
+		Root (CONCAT(v,__root__)); \
+		(CONCAT(v,__root__)).p = (void **) &v; \
+		(CONCAT(v,__root__)).next = rootlist; \
+		rootlist = &(CONCAT(v,__root__))
 #define	RefPop(v) \
-		assert(rootlist == &(__root__ ## v)); \
+		assert(rootlist == &(CONCAT(v,__root__))); \
 		assert(rootlist->p == (void **) &v); \
 		rootlist = rootlist->next;
 #define RefEnd(v) \
-		assert(rootlist == &(__root__ ## v)); \
-		assert(rootlist->p == (void **) &v); \
 		RefPop(v); \
-	} while (0)
-#define RefReturn(v)	/* { */	\
-		assert(rootlist == &(__root__ ## v)); \
-		assert(rootlist->p == (void **) &v); \
+	}
+#define RefReturn(v) \
 		RefPop(v); \
 		return v; \
-	} while (0)
+	}
+#define	RefAdd(e) \
+	if (0) ; else { \
+		Root __root__; \
+		__root__.p = (void **) &e; \
+		__root__.next = rootlist; \
+		rootlist = &__root__
+#define	RefRemove(e) \
+		assert(rootlist == &__root__); \
+		assert(rootlist->p == (void **) &e); \
+		rootlist = rootlist->next; \
+	}
 
 #define	RefEnd2(v1, v2)		RefEnd(v1); RefEnd(v2)
 #define	RefEnd3(v1, v2, v3)	RefEnd(v1); RefEnd2(v2, v3)
@@ -468,18 +400,17 @@ struct Handler {
 	jmp_buf label;
 };
 
-extern Handler childhandler, *bottomhandler;
+extern Handler childhandler, *tophandler;
 extern List *exception;
 extern List *pushhandler(Handler *handler);	/* must be a macro */
 extern void pophandler(Handler *handler);
 extern noreturn throw(List *exc);
-extern noreturn fail(const char *name, ...);
+extern noreturn fail(const char *name VARARGS);
 extern void newchildcatcher(void);
 
 #define	pushhandler(hp)	( \
 		((hp)->rootlist = rootlist), \
-		((hp)->up = bottomhandler), \
-		(bottomhandler = (hp)), \
+		((hp)->up = tophandler), \
+		(tophandler = (hp)), \
 		(setjmp((hp)->label) ? exception : NULL) \
 	)
-
