@@ -5,26 +5,24 @@
 
 /* globals */
 Handler childhandler = { NULL, NULL };
-Handler *bottomhandler = NULL;
+Handler *tophandler = NULL;
 List *exception = NULL;
 
 /* pophandler -- remove a handler */
 extern void pophandler(Handler *handler) {
-	assert(bottomhandler == handler);
+	assert(tophandler == handler);
 	assert(handler->rootlist == rootlist);
-	bottomhandler = handler->up;
+	tophandler = handler->up;
 }
 
 /* throw -- raise an exception */
 extern noreturn throw(List *exc) {
-	Handler *handler = bottomhandler;
+	Handler *handler = tophandler;
 
-	debug("<< throwing from %ux : %L >>\n", &handler, exc, ", ");
-
-	assert(gcblocked == 0);
+	assert(!gcisblocked());
 	assert(exc != NULL);
 	assert(handler != NULL);
-	bottomhandler = handler->up;
+	tophandler = handler->up;
 	for (; rootlist != handler->rootlist; rootlist = rootlist->next)
 		assert(rootlist != NULL);
 	exception = exc;
@@ -32,24 +30,25 @@ extern noreturn throw(List *exc) {
 }
 
 /* fail -- pass a user catchable error up the exception chain */
-extern noreturn fail(const char *fmt, ...) {
+extern noreturn fail VARARGS1(const char *, fmt) {
 	char *s;
 	va_list args;
 
-	va_start(args, fmt);
+	VA_START(args, fmt);
 	s = strv(fmt, args);
 	va_end(args);
 
 	gcdisable(0);
 	Ref(List *, e, mklist(mkterm("error", NULL),
 			      mklist(mkterm(s, NULL), NULL)));
-	gcenable();
+	while (gcisblocked())
+		gcenable();
 	throw(e);
 	RefEnd(e);
 }
 
 /* newchildcatcher -- remove the current handler chain for a new child */
 extern void newchildcatcher(void) {
-	bottomhandler = &childhandler;
+	tophandler = &childhandler;
 	/* TODO: flush some of the rootlist? */
 }
