@@ -1,4 +1,4 @@
-/* heredoc.c -- in-line files (here documents) ($Revision: 1.1.1.1 $) */
+/* heredoc.c -- in-line files (here documents) */
 
 #include "es.h"
 #include "gc.h"
@@ -17,13 +17,11 @@ static Here *hereq;
 extern Tree *getherevar(void) {
 	int c;
 	char *s;
-	size_t len;
 	Buffer *buf = openbuffer(0);
 	while (!dnw[c = GETC()])
 		buf = bufputc(buf, c);
-	len = buf->len;
 	s = sealcountedbuffer(buf);
-	if (len == 0) {
+	if (buf->len == 0) {
 		yyerror("null variable name in here document");
 		return NULL;
 	}
