@@ -1,4 +1,4 @@
-/* input.h -- definitions for es lexical analyzer ($Revision: 1.3 $) */
+/* input.h -- definitions for es lexical analyzer */
 
 #define	MAXUNGET	2		/* maximum 2 character pushback */
 
@@ -14,7 +14,11 @@ struct Input {
 	int ungot;
 	int lineno;
 	int fd;
-	int runflags;
+	Boolean interactive;
+	Boolean echoinput;
+#if LISPTREES
+	Boolean lisptrees;
+#endif
 };
 
 
