@@ -7,7 +7,7 @@
  * allocation and garbage collector support
  */
 
-static Tag ListTag;
+DefineTag(List, static);
 
 extern List *mklist(Term *term, List *next) {
 	gcdisable(0);
@@ -32,8 +32,6 @@ static size_t ListScan(void *p) {
 	return sizeof (List);
 }
 
-static DefineTag(List);
-
 
 /*
  * basic list manipulations
@@ -54,7 +52,7 @@ extern List *safereverse(List *list) {
 	for (lp = NULL; list != NULL; list = list->next)
 		lp = mklist(list->term, lp);
 
-	Ref(List*, result, lp);
+	Ref(List *, result, lp);
 	gcenable();
 	RefReturn(result);
 }
@@ -75,23 +73,29 @@ extern List *reverse(List *list) {
 }
 
 /* append -- merge two lists, non-destructively */
-extern List *append(List *head0, List *tail0) {
-	List **prevp;
+extern List *append(List *head, List *tail) {
+	List *lp, **prevp;
 #if 0	/* if you want this optimization, rewrite listcopy */
 	if (tail0 == NULL)
 		return head0;
 #endif
-	Ref(List *, lp, NULL);
-	Ref(List *, head, head0);
-	Ref(List *, tail, tail0);
+	Ref(List *, hp, head);
+	Ref(List *, tp, tail);
+	gcdisable(40 * sizeof (List));
+	head = hp;
+	tail = tp;
+	RefEnd2(tp, hp);
+
 	for (prevp = &lp; head != NULL; head = head->next) {
 		List *np = mklist(head->term, NULL);
 		*prevp = np;
 		prevp = &np->next;
 	}
 	*prevp = tail;
-	RefEnd2(tail, head);
-	RefReturn(lp);
+
+	Ref(List *, result, lp);
+	gcenable();
+	RefReturn(result);
 }
 
 /* listcopy */
@@ -126,3 +130,17 @@ extern Term *nth(List *list, int n) {
 	}
 	return NULL;
 }
+
+/* sortlist */
+extern List *sortlist(List *list) {
+	if (length(list) > 1) {
+		Vector *v = vectorize(list);
+		sortvector(v);
+		gcdisable(0);
+		Ref(List *, lp, listify(v->count, v->vector));
+		gcenable();
+		list = lp;
+		RefEnd(lp);
+	}
+	return list;
+}
