@@ -1,19 +1,17 @@
 /* open.c -- to insulate <fcntl.h> from the rest of rc */
 
-#include <fcntl.h>
+#define	REQUIRE_FCNTL	1
+
 #include "es.h"
-#include "token.h"
 
 /* prototype for open() follows. comment out if necessary */
 #if __MACH__
-extern int open(char *, int, ...);
+extern int open(char *, int VARARGS);
 #endif
 
 /*
  * Opens a file with the necessary flags.  Assumes the following order:
- *	typedef enum {
- *		rOpen, rCreate, rAppend, rDup, rClose, rPipe, rHeredoc, rHerestring
- *	} RedirKind;
+ *	typedef enum { oOpen, oCreate, oAppend } OpenKind;
  */
 
 static int mode_masks[] = {
@@ -22,7 +20,7 @@ static int mode_masks[] = {
 	O_APPEND | O_CREAT | O_WRONLY	/* rAppend */
 };
 
-extern int eopen(char *name, RedirKind k) {
+extern int eopen(char *name, OpenKind k) {
 	assert((unsigned) k < arraysize(mode_masks));
 	return open(name, mode_masks[k], 0666);
 }
