@@ -6,15 +6,15 @@
 Boolean sigint_newline = TRUE;
 
 jmp_buf slowlabel;
-volatile SIG_ATOMIC_T slow = FALSE;
-volatile SIG_ATOMIC_T interrupted = FALSE;
-static void (*sighandlers[NUMOFSIGNALS])(int);
-static volatile SIG_ATOMIC_T sigcount;
-static volatile SIG_ATOMIC_T caught[NUMOFSIGNALS];
+Atomic slow = FALSE;
+Atomic interrupted = FALSE;
+static sigresult (*sighandlers[NUMOFSIGNALS])(int);
+static Atomic sigcount;
+static Atomic caught[NUMOFSIGNALS];
 
-static void (*default_sigint)(int) = SIG_DFL,
-	    (*default_sigquit)(int) = SIG_DFL,
-	    (*default_sigterm)(int) = SIG_DFL;
+static sigresult (*default_sigint)(int)  = SIG_DFL,
+		 (*default_sigquit)(int) = SIG_DFL,
+		 (*default_sigterm)(int) = SIG_DFL;
 
 /* catcher -- the signal handler routine */
 extern void catcher(int sig) {
@@ -35,7 +35,7 @@ extern void catcher(int sig) {
 extern void sigchk(void) {
 	if (sigcount != 0) {
 		int sig;
-		void (*h)(int);
+		sigresult (*h)(int);
 		if (hasforked)
 			exit(1); /* exit unconditionally on a signal in a child process */
 		for (sig = 0;; sig++) {
@@ -55,8 +55,8 @@ extern void sigchk(void) {
 	}
 }
 
-extern void (*esignal(int sig, void (*h)(int)))(int) {
-	void (*old)(int);
+extern sigresult (*esignal(int sig, sigresult (*h)(int)))(int) {
+	sigresult (*old)(int);
 	SIGCHK();
 	old = sighandlers[sig];
 	if (h == SIG_DFL || h == SIG_IGN) {
@@ -69,35 +69,57 @@ extern void (*esignal(int sig, void (*h)(int)))(int) {
 	return old;
 }
 
-static void sigint(int sig) {
+static sigresult sigint(int sig) {
 	assert(sig == SIGINT);
 	/* this is the newline you see when you hit ^C while typing a command */
 	if (sigint_newline)
 		eprint("\n");
 	sigint_newline = TRUE;
-	/* TODO: redirq = NULL;	cond = FALSE; */
-	while (gcblocked)
+	while (gcisblocked())
 		gcenable();
 	throw(mklist(mkterm("signal", NULL), mklist(mkterm("sigint", NULL), NULL)));
+	unreached(SIGRESULT);
 }
 
-static void noop(int sig) {
+static sigresult trapsig(int sig) {
+	while (gcisblocked())
+		gcenable();
+	throw(mklist(mkterm("signal", NULL), mklist(mkterm(signals[sig].name, NULL), NULL)));
+	unreached(SIGRESULT);
+}
+
+static sigresult noop(int sig) {
+	return SIGRESULT;
+}
+
+extern void trapsignals(Boolean sigs[NUMOFSIGNALS]) {
+	int sig;
+	for (sig = 1; sig < NUMOFSIGNALS; sig++)
+		if (sigs[sig]) {
+#if SPECIAL_SIGCLD
+			if (sig == SIGCLD)
+				fail("cannot trap SIGCLD");
+#endif
+			if (sighandlers[sig] != trapsig)
+				esignal(sig, trapsig);
+		} else if (sighandlers[sig] == trapsig)
+			esignal(sig, SIG_IGN);
 }
 
 extern void initsignals(Boolean allowdumps) {
 	int sig;
 	for (sig = 1; sig < NUMOFSIGNALS; sig++) {
-		void (*h)(int) = signal(sig, SIG_DFL);
+		sigresult (*h)(int) = signal(sig, SIG_DFL);
 		if (h != SIG_DFL)
 			signal(sig, h);
 		sighandlers[sig] = h;
 	}
-	if (interactive || sighandlers[SIGINT] != SIG_IGN)
+	if (isinteractive() || sighandlers[SIGINT] != SIG_IGN)
 		esignal(SIGINT, default_sigint = sigint);
 	if (!allowdumps) {
-		if (interactive)
+		if (isinteractive())
 			esignal(SIGTERM, default_sigterm = noop);
-		if (interactive || sighandlers[SIGQUIT] != SIG_IGN)
+		if (isinteractive() || sighandlers[SIGQUIT] != SIG_IGN)
 			esignal(SIGQUIT, default_sigquit = noop);
 	}
 }
@@ -111,13 +133,13 @@ extern Boolean issilentsignal(List *e) {
 extern void setsigdefaults(Boolean background) {
 
 #if JOB_PROTECT && SIGTTOU && SIGTTIN && SIGTSTP
-#define	IGNORE(signal) ((default_ ## signal) = SIG_DFL)
+#define	IGNORE(signal) (CONCAT(default_,signal) = SIG_DFL)
 #else
 #define	IGNORE(signal) \
 	if (!background) \
-		(default_ ## signal) = SIG_DFL; \
+		CONCAT(default_,signal) = SIG_DFL; \
 	else { \
-		(default_ ## signal) = SIG_IGN; \
+		CONCAT(default_,signal) = SIG_IGN; \
 		esignal(sig, SIG_IGN); \
 		/* TODO: remove #signal from $signals */ \
 		break; \
@@ -125,7 +147,6 @@ extern void setsigdefaults(Boolean background) {
 #endif
 
 	int sig;
-	closefds();
 	for (sig = 1; sig < NUMOFSIGNALS; sig++)
 		if (sighandlers[sig] != SIG_IGN) {
 			sighandlers[sig] = NULL;
@@ -147,5 +168,4 @@ extern void setsigdefaults(Boolean background) {
 				}
 			}
 		}
-	/* TODO: sigexit */
 }
