@@ -3,9 +3,6 @@
 #include "es.h"
 #include "sigmsgs.h"
 
-extern long strtol(const char *nptr, char **endptr, int base);
-
-
 static Term trueterm = { "0", NULL }, falseterm = { "1", NULL };
 static List truelist = { &trueterm, NULL }, falselist = { &falseterm, NULL };
 List *true = &truelist, *false = &falselist;
@@ -66,13 +63,13 @@ extern void printstatus(int pid, int status) {
 	if (status & 0xff) {
 		char *msg = ((status & 0x7f) < NUMOFSIGNALS ? signals[status & 0x7f].msg : "");
 		if (pid != 0)
-			print("%d: ", pid);
+			eprint("%d: ", pid);
 		if (status & 0x80) {
 			if (*msg == '\0')
-				print("core dumped\n");
+				eprint("core dumped\n");
 			else
-				print("%s--core dumped\n", msg);
+				eprint("%s--core dumped\n", msg);
 		} else if (*msg != '\0')
-			print("%s\n", msg);
+			eprint("%s\n", msg);
 	}
 }
