@@ -1,16 +1,18 @@
-/* syntax.c -- abstract syntax tree re-writing rules ($Revision: 1.3 $) */
+/* syntax.c -- abstract syntax tree re-writing rules */
 
 #include "es.h"
 #include "input.h"
 #include "syntax.h"
 #include "token.h"
 
-Tree errornode;
-Tree *parsetree;
+Tree *ifs, *parsetree;
 
 /* initparse -- called at the dawn of time */
 extern void initparse(void) {
 	globalroot(&parsetree);
+	globalroot(&ifs);
+
+	ifs = mk(nVar, mk(nWord, "ifs"));
 }
 
 /* treecons -- create new tree list cell */
@@ -163,7 +165,7 @@ extern Tree *redirect(Tree *t) {
 	}
 	if (firstis(r, "%heredoc"))
 		if (!queueheredoc(r))
-			return &errornode;
+			return NULL;
 	p->CAR = thunkify(redirect(t));
 	return r;
 }
