@@ -1,4 +1,4 @@
-/* syntax.h -- abstract syntax tree interface ($Revision: 1.3 $) */
+/* syntax.h -- abstract syntax tree interface */
 
 #define	CAR	u[0].p
 #define	CDR	u[1].p
@@ -6,8 +6,6 @@
 
 /* syntax.c */
 
-extern Tree errornode;
-
 extern Tree *treecons(Tree *car, Tree *cdr);
 extern Tree *treecons2(Tree *car, Tree *cdr);
 extern Tree *treeconsend(Tree *p, Tree *q);
