@@ -3,7 +3,7 @@
 #include "es.h"
 #include "gc.h"
 
-static Tag TermTag;
+DefineTag(Term, static);
 
 extern Term *mkterm(char *str, Closure *closure) {
 	gcdisable(0);
@@ -14,18 +14,6 @@ extern Term *mkterm(char *str, Closure *closure) {
 	RefReturn(term);
 }
 
-extern char *getstr(Term *term) {
-	assert(term != NULL);
-	Ref(Term *, tp, term);
-	if (tp->str == NULL) {
-		assert(tp->closure != NULL);
-		tp->str = str("%C", tp->closure);
-	}
-	term = tp;
-	RefEnd(tp);
-	return term->str;
-}
-
 extern Closure *getclosure(Term *term) {
 	if (term->closure == NULL) {
 		char *s = term->str;
@@ -33,6 +21,7 @@ extern Closure *getclosure(Term *term) {
 		if (
 			((*s == '{' || *s == '@') && s[strlen(s) - 1] == '}')
 			|| (*s == '$' && s[1] == '&')
+			|| hasprefix(s, "%closure")
 		) {
 			Ref(Term *, tp, term);
 			Ref(Tree *, np, parsestring(s));
@@ -41,6 +30,7 @@ extern Closure *getclosure(Term *term) {
 				return NULL;
 			}
 			tp->closure = extractbindings(np);
+			tp->str = NULL;
 			term = tp;
 			RefEnd2(np, tp);
 		}
@@ -71,9 +61,8 @@ static void *TermCopy(void *op) {
 
 static size_t TermScan(void *p) {
 	Term *term = p;
-	term->str = forward(term->str);
 	term->closure = forward(term->closure);
+	term->str = forward(term->str);
 	return sizeof (Term);
 }
 
-static DefineTag(Term);
